import sys
import ctypes
import inspect

flags = sys.getdlopenflags()
sys.setdlopenflags(flags | ctypes.RTLD_GLOBAL)

# Dig past the extra python wrapping to the direct ScannerBit.so shared library interface level
import pyscannerbit.scan as sb
from pyscannerbit.ScannerBit.python import ScannerBit

# define likelihood, technically optional
def like(m):
    #print("m:", m)
    #print("m.items():", [(k,v) for k,v in m.items()])
    a = m["model1::x"]
    ScannerBit.print("my_param", 0.5) # can print custom parameters 

    return -a*a/2.0

def like_a(a,b):
    ScannerBit.print("a", a) # can print custom parameters
    ScannerBit.print("b", b) # can print custom parameters
    return -a*a/2.0

def like_x(x,y):
    ScannerBit.print("x", x) # can print custom parameters
    ScannerBit.print("y", y) # can print custom parameters
    return -x*y/2.0
     
# Try to dynamically generate likelihood function
def wrap_function():
    def wrapped_function(m):
        #print("m:", m)
        a = m["model1::x"]
        b = 3
        return like_a(a,b) 
    return wrapped_function

# Do it from a class
class Test:
    def __init__(self,f):
        self.f_in = f
        self.f = self._wrap_function()
        signature = inspect.getargspec(self.f_in)
        self._argument_names = signature.args
        self._model_name = "model1"
        print("self._argument_names:", self._argument_names)
 
    def _wrap_function(self):
        def wrapped_function(par_dict):
            print("par_dict:", par_dict)
            print("par_dict.items():", [(k,v) for k,v in par_dict.items()])
            arguments = [par_dict["{}::{}".format(self._model_name, n)]
              for n in self._argument_names]
            return self.f_in(*arguments, **(self.kwargs or {}))
        return wrapped_function


#like_dyn = wrap_function()
t = Test(like_x)
like_dyn = t.f 

print("like_dyn:",like_dyn)
print("args: ",inspect.getargspec(like_dyn))

# Version where ScannerBit interface is passed through via the wrapper
# (needed to allow multiple scans per python session)
def like2(scan,m):
    a = m["model1::x"]
    scan.print("my_param", 0.5) # can print custom parameters 
    return -a*a/2.0


# define prior, optional
def prior(vec, map):
    # tell ScannerBit that the hypergrid dimension is 1
    ScannerBit.ensure_size(vec, 1) # this needs to be the first line!

    map["model1::x"] = 5.0 - 10.0*vec[0]

# declare scan object
myscan = ScannerBit.scan(True)

settings = {
"Parameters": {
  "model1": {
    "x": None,
    "y": None,
    }
  },
"Priors": {
  "x_prior": {
    "prior_type": 'flat',
    "parameters": ['model1::x'],
    "range": [1.0, 40.0],
    },
  "y_prior": {
    "prior_type": 'flat',
    "parameters": ['model1::y'],
    "range": [1.0, 40.0],
    }
  },
"Printer": {
  "printer": "hdf5",
  "options": {
    "output_file": "results.hdf5",
    "group": "/",
    "delete_file_on_restart": True,
    }
  },
"Scanner": {
  "scanners": {
    "twalk": {
      "plugin": "twalk",
      "like": "LogLike",
      "tolerance": 1.003,
      "kwalk_ratio": 0.9,
      "projection_dimension": 4
      }
    },
  "use_scanner": "twalk",
  },
"KeyValues": {
  "default_output_path": "pyscannerbit_run_data/",
  "likelihood": {
    "model_invalid_for_lnlike_below": -1e6
    }
  }
}

#myscan.run(inifile=settings, lnlike={"LogLike": like}, prior=prior, restart=True)

# Try via wrapper to the above interface:
#sb._run_scan(settings, like, prior)

# Try with dummy prior
def dummy_prior(vec, map):
    ScannerBit.ensure_size(vec, 2) # this needs to be the first line!
    map["model1::x"] = 5.0 - 10.0*vec[0]
    map["model1::y"] = 5.0 - 10.0*vec[1]

myscan.run(inifile=settings, lnlike={"LogLike": like}, prior=dummy_prior, restart=True)

# Try with dynamically created function
#sb._run_scan(settings, like_dyn, dummy_prior)

#Try without prior
#sb._run_scan(settings, like2, "") # Fails!
 



import copy

# Nested dictionary merge function
# Modified from: https://stackoverflow.com/a/7205107/1447953 
def _merge(a, b, path=None):
    "merges b into a"
    if path is None: path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                _merge(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass # same leaf value
            else:
                #raise Exception('Conflict at %s' % '.'.join(path + [str(key)]))
                # Actually the behaviour we want is for a to override b, so do nothing here
                pass
        else:
            a[key] = copy.deepcopy(b[key])
    return a




"""
Class representing HDF5 results from ScannerBit
===============================================
"""

import h5py
import matplotlib.pyplot as plt
import numpy as np


class HDF5(h5py.File):
    """
    """
    def __init__(self, file_name, model=None, loglike="LogLike", posterior="Posterior"):
        """
        """
        self.loglike = loglike
        self.posterior = posterior
        super(HDF5, self).__init__(file_name, mode='r')
        self.group = self.keys()[0]
        self.model = model if model else self.get_model_name()

    def get_model_name(self):
        """
        """
        for k in self[self.group].keys():
            if "::" in k:
                return k.split("::")[0]

    def get_param_names(self):
        """
        """
        prefix = "{}::".format(self.model)
        suffix = "_isvalid"
        return [k[len(prefix):] for k in self[self.group].keys()
            if k.startswith(prefix) and not k.endswith(suffix)]

    def get_param(self, name):
        """
        """
        try:
            valid = self[self.group]["{}_isvalid".format(name)]
        except KeyError:
            name = "{}::{}".format(self.model, name)
            valid = self[self.group]["{}_isvalid".format(name)]

        mask = np.array(valid, dtype=np.bool)
        return self[self.group][name][mask]

    def get_loglike(self):
        """
        """
        return self.get_param(self.loglike)

    def get_posterior(self):
        """
        """
        return self.get_param(self.posterior)

    def get_best_fit(self, name):
        """
        """
        loglike = self.get_loglike()
        index = np.argmax(loglike)
        return self.get_param(name)[index]

    def get_min_chi_squared(self):
        """
        """
        return -2. * self.get_loglike().max()

    def make_plot(self, name_x, name_y):
        """
        """
        fig = plt.figure(figsize=(8, 6))
        ax = fig.add_subplot(111)

        ax.scatter(self.get_param(name_x), self.get_param(name_y),
          marker="o", facecolor='k', edgecolor='', alpha=0.5, s=20)
        ax.scatter(self.get_best_fit(name_x), self.get_best_fit(name_y),
          marker="*", facecolor='Gold', alpha=1., s=250, label="Best-fit")

        ax.set_xlabel(name_x)
        ax.set_ylabel(name_y)
        ax.legend()
        plt.show()



"""
Modified PyYAML
===============

Supports !import keyword, lists as keys and checks for duplicate keys.
"""

import os
import yaml
from yaml import dump


IMPORT_TAG = '!import'
SEQ_TAG = 'tag:yaml.org,2002:seq'


class ImportLoader(yaml.Loader):
    """
    Override default loader to support inclusion of YAML files and permit lists
    as YAML keys.
    """
    def __init__(self, stream):
        """
        Obtain path to stream, if possible.
        """
        try:
            self._root = os.path.split(stream.name)[0]
        except AttributeError:
            self._root = os.getcwd()
        self.duplicates = []
        super(ImportLoader, self).__init__(stream)

    def import_(self, node):
        """
        Import YAML file via a keyword (see
        https://stackoverflow.com/a/9577670/2855071)

        :returns: YAML file for inclusion
        :rtype: dict
        """
        yamal_name = os.path.join(self._root, self.construct_scalar(node))

        with open(yamal_name, 'r') as yamal_file:
            return yaml.load(yamal_file, ImportLoader)

    def construct_tuple(self, node):
        """
        Permit lists as YAML keys (see
        https://gist.github.com/miracle2k/3184458#file-tuple-py).
        """
        return tuple(ImportLoader.construct_sequence(self, node))

    def note_duplicates(self, node, deep=False):
        """
        Raise error if duplicate keys (see
        https://gist.github.com/pypt/94d747fe5180851196eb).
        """
        mapping = {}

        for key_node, value_node in node.value:
            key = self.construct_object(key_node, deep=deep)
            value = self.construct_object(value_node, deep=deep)
            if key in mapping:
                self.duplicates.append(key)
            mapping[key] = value

        return self.construct_mapping(node, deep)

ImportLoader.add_constructor(yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, ImportLoader.note_duplicates)
ImportLoader.add_constructor(SEQ_TAG, ImportLoader.construct_tuple)
ImportLoader.add_constructor(IMPORT_TAG, ImportLoader.import_)


def load(stream):
    """
    :param stream: YAML file to be loaded
    :type stream: Stream obj. File or str

    :returns: YAML file as nested dictionary
    :rtype: dict
    """
    return yaml.load(stream, ImportLoader)

def duplicates(stream):
    """
    """
    loader = ImportLoader(stream)
    try:
        loader.get_single_data()
        return loader.duplicates
    finally:
        loader.dispose()



"""Helper routines for accessing HDF5 output of ScannerBit"""

import h5py
import matplotlib.pyplot as plt
import numpy as np
import pyscannerbit.plottools as pt

# Helper class to manage data/is_valid HDF5 dataset pairs 
class DataPair:
  def __init__(self,g,name,model=None):
     try:
         self._valid = g["{}_isvalid".format(name)]
     except KeyError:
         name = "{}::{}".format(model, name)
         self._valid = g["{}_isvalid".format(name)]
     self._data  = g[name]
  
  def data(self):
     return self._data

  def valid(self):
     return np.array(self._valid, dtype=np.bool_)

  def validdata(self):
     return self.data()[self.valid()]


def get_data(in_group, hdf5_dataset_names, apply_common_mask=True, model=None):
  datapairs = []
  for name in hdf5_dataset_names:
    datapairs += [ DataPair(in_group,name,model) ]
  output = [] 
  if apply_common_mask:
     m = datapairs[0].valid()
     for d in datapairs[1:]:
        m = m & d.valid()
     for d in datapairs:
        output += [d.data()[m]]
  else:
     output = datapairs
  return output


class HDF5(h5py.Group):
    """
    Class representing HDF5 results from ScannerBit
    ===============================================
    """
    def __init__(self, group, model=None, loglike="LogLike", posterior="Posterior", h5file=None):
        """Wrap a group in a HDF5 file
        """
        self.loglike = loglike
        self.posterior = posterior
        self.h5file = h5file
        super(HDF5,self).__init__(group)
        self.model = model if model else self.get_model_name()

    @classmethod
    def fromFile(clsobj, h5filename, group, model=None, loglike="LogLike", posterior="Posterior"):
        """Alternate constructor to wrap a group directly from a file, rather than having to
           open the file before constructing this object.
        """
        f = h5py.File(h5filename,'r')
        g = f[group] 
        return clsobj(g,model,loglike,posterior,h5file=f)

    def get_model_name(self):
        """Attempt to infer model name from parameter dataset label
        """
        for k in self.keys():
            if "::" in k:
                return k.split("::")[0]

    def get_param_names(self):
        """
        """
        prefix = "{}::".format(self.model)
        suffix = "_isvalid"
        return [k[len(prefix):] for k in self.keys()
            if k.startswith(prefix) and not k.endswith(suffix)]

    def get_params(self, names):
        return get_data(self, names, model=self.model, apply_common_mask=True)

    def get_param(self, name):
        return get_data(self, [name], model=self.model, apply_common_mask=True)[0]

    def get_loglike(self):
        """
        """
        return self.get_param(self.loglike)

    def get_posterior(self):
        """
        """
        return self.get_param(self.posterior)

    def get_best_fit(self, name):
        """
        """
        loglike = self.get_loglike()
        index = np.argmax(loglike)
        return self.get_param(name)[index]

    def get_min_chi_squared(self):
        """
        """
        return -2. * self.get_loglike().max()

    def make_plot(self, name_x, name_y):
        """
        """
        fig = plt.figure(figsize=(8, 6))
        ax = fig.add_subplot(111)

        ax.scatter(self.get_param(name_x), self.get_param(name_y),
          marker="o", facecolor='k', edgecolor='', alpha=0.5, s=20)
        ax.scatter(self.get_best_fit(name_x), self.get_best_fit(name_y),
          marker="*", facecolor='Gold', alpha=1., s=250, label="Best-fit")

        ax.set_xlabel(name_x)
        ax.set_ylabel(name_y)
        ax.legend()
        plt.show()

    def plot_profile_likelihood(self,ax,xpar,ypar,use_default_model_name=True,nxbins=50,nybins=50):
        if(use_default_model_name):
           xpar = "{0}::{1}".format(self.model,xpar)
           ypar = "{0}::{1}".format(self.model,ypar)
        logl,x,y = get_data(self,[self.loglike,xpar,ypar])
        data = np.vstack((x,y,-2*logl)).T
        pt.profplot(ax,data,title=None,labels=[xpar,ypar],nxbins=nxbins,nybins=nybins)

    def plot_marginal_posterior(self,ax,xpar,ypar,use_default_model_name=True,nxbins=100,nybins=100):
        if(use_default_model_name):
           xpar = "{0}::{1}".format(self.model,xpar)
           ypar = "{0}::{1}".format(self.model,ypar)
        p,x,y = get_data(self,[self.posterior,xpar,ypar])
        data = np.vstack((x,y,p)).T
        pt.margplot(ax,data,title=None,labels=[xpar,ypar],nxbins=nxbins,nybins=nybins)




"""
Pythonic interface to ScannerBit
================================
"""

import sys
import ctypes
import inspect

sys.setdlopenflags(sys.getdlopenflags() | ctypes.RTLD_GLOBAL)

import pyscannerbit
import default
import import_yaml
from hdf5_reader import HDF5


DIR = "runs/test_scan/samples/"


class Scan(object):
    """
    """
    def __init__(self, function, bounds, prior_types=None, kwargs=None, scanner="multinest",
      settings=None, yaml="scan.yaml"):
        """
        """
        self.function = function
        self.bounds = bounds
        self.prior_types = prior_types if prior_types else ["flat"] * len(bounds)
        self.scanner = scanner
        self.settings = settings if settings else default.settings
        self.yaml = yaml
        self.kwargs = kwargs

        signature = inspect.getargspec(self.function)
        n_kwargs = len(signature.defaults or [])
        self._argument_names = signature.args[:-n_kwargs or None]
        assert len(self._argument_names) == len(self.bounds)

        self._model_name = self.settings["Parameters"].keys()[0]
        self._wrapped_function = self._wrap_function()
        self._scanned = False
        self._write_yaml()

    def _get_hdf5_name(self):
        """
        """
        assert self.settings["Printer"]["printer"] == "hdf5"
        file_name = self.settings["Printer"]["options"]["output_file"]
        return "{}/{}".format(DIR, file_name)

    def _write_yaml(self):
        """
        """
        self.settings["Scanner"]["use_scanner"] = self.scanner
        self.settings["Parameters"][self._model_name] = dict()
        for n, b, t in zip(self._argument_names, self.bounds, self.prior_types):
            self.settings["Parameters"][self._model_name][n] = {"range": b, "prior_type": t}

        with open(self.yaml, 'w') as f:
            import_yaml.dump(self.settings, stream=f)

    def _wrap_function(self):
        """
        """
        def wrapped_function(par_dict):
            """
            """
            arguments = [par_dict["{}::{}".format(self._model_name, n)]
              for n in self._argument_names]
            return self.function(*arguments, **(self.kwargs or {}))

        return wrapped_function

    def scan(self):
        """
        """
        pyscannerbit.run_scan(self.yaml, self._wrapped_function)
        self._scanned = True

    def get_hdf5(self):
        """
        """
        assert self._scanned
        return HDF5(self._get_hdf5_name(), model=self._model_name)



"""Demo script which runs all the (serious) scanners to which pyScannerBit has access"""

import os
import numpy as np
import math
import matplotlib.pyplot as plt
import h5py
from pyscannerbit.hdf5_help import get_data, HDF5 
from pyscannerbit import ext_module as ext
from mpi4py import MPI
rank = MPI.COMM_WORLD.Get_rank()
size = MPI.COMM_WORLD.Get_size()

# Regenerate scan data?
new_scans = True

# Test function
def rastrigin(m):
    X = [m["model::x"],m["model::y"],m["model::z"]]
    A = 10
    return - (A + sum([(x**2 - A * np.cos(2 * math.pi * x)) for x in X]))

# Prior transformation from unit hypercube
def prior(vec, map):
    vec.ensure_size(3)
    map["model::x"] = -4 + 8*vec[0] # flat prior over [-4,4]
    map["model::y"] = -4 + 8*vec[1]
    map["model::z"] = -4 + 8*vec[2]

# Settings for quick and dirty scans. Won't do very well, because the test function is
# actually rather tough!
# Don't have to specify all scanner options; anything missing will revert to defaults (see defaults.py)
settings = {
"Parameters": {
  "model1": {
    "x": None,
    }
  },
"Priors": {
  "x_prior": {
    "prior_type": 'flat',
    "parameters": ['model1::x'],
    "range": [1.0, 40.0],
    }
  },
"Printer": {
  "printer": "hdf5",
  "options": {
    "output_file": "results.hdf5",
    "group": "/",
    "delete_file_on_restart": "true",
    }
  },
"Scanner": {
  "scanners": {
    "multinest": {"plugin": "multinest", "like": "LogLike", "tol": 0.5, "nlive": 100},
    "twalk": {"plugin": "twalk", "like": "LogLike", "sqrtR": 1.05},
    "polychord": {"plugin": "polychord", "like": "LogLike", "tol": 1.0, "nlive": 20},
    "diver": {"plugin": "diver", "like": "LogLike", "convthresh": 1e-2, "NP": 300}, 
    "random": {"plugin": "random", "like": "LogLike", "point_number": 10000},
    "toy_mcmc": {"plugin": "toy_mcmc", "like": "LogLike", "point_number": 10}, # Acceptance ratio is really bad with this scanner, so don't ask for much
    "badass": {"plugin": "badass", "like": "LogLike", "points": 1000, "jumps": 10},
    "pso": {"plugin": "pso", "like": "LogLike", "NP": 400}
    },    
  "use_scanner": "multinest",
  "objectives": {
    "gaussian": {
      "plugin": "gaussian",
      "purpose": "LogLike",
      "parameters": {
        "param...20": None,
        "range": [-5, 5]
        }
      }
    }
  },
"KeyValues": {
  "default_output_path": "pyscannerbit_run_data/",
  "likelihood": {
    "model_invalid_for_lnlike_below": -1e6
    }
  }
}
  
def _get_hdf5_group(settings):
    """
    """
    assert settings["Printer"]["printer"] == "hdf5"
    file_name = settings["Printer"]["options"]["output_file"]
    group_name = settings["Printer"]["options"]["group"]
    DIR = settings["KeyValues"]["default_output_path"]
    fullpath = "{}/samples/{}".format(DIR, file_name)
    f = h5py.File(fullpath,'r')
    g = f[group_name]
    return g

def get_hdf5(settings):
    
    loglike_par = "LogLike"
    scanner = settings["Scanner"]["use_scanner"] # Might have been None and then filled by _process_settings with defaults
    if scanner == "multinest" or scanner == "polychord":
        posterior_par = "Posterior"
    elif scanner == "twalk":
        posterior_par = "mult"
    else:
        posterior_par = None
    g = HDF5(_get_hdf5_group(settings).id,model="model",
                loglike=loglike_par, posterior=posterior_par)
    return g

def rm_samples(settings):
    file_name = settings["Printer"]["options"]["output_file"]
    DIR = settings["KeyValues"]["default_output_path"]
    fullpath = "{}/samples/{}".format(DIR, file_name)
    try:
        os.remove(fullpath)
    except:
        print("No such file: " + fullpath)


scanners = ["multinest","polychord","diver","twalk"]
#colors = ["r","m","b","g"]
#scanners = ["twalk","badass","pso"]
colors = ["r","b","g"]
if size is 1:
    scanners += ["random","toy_mcmc"] # "random" and "toy_mcmc" seem to not be MPI compatible. Should make GAMBIT throw an error about this, or fix them.
    colors += ["c","y"]

# Test just one scanner
#scanners = ["pso"]
#colors = ["r"]

Nscans = len(scanners)
results = {}

# Do all scans
for s in scanners:
    # Create scan manager object
    # (prior_types argument currently does nothing)
    # myscan = sb.Scan(rastrigin, prior_func=prior, scanner=s, scanner_options=scanner_options[s])
    rm_samples(settings)
    myscan = ext.sb.scan(False)
    settings["Scanner"]["use_scanner"] = s
    if new_scans:
        print("Running scan with {}".format(s))
        myscan.run(inifile=settings, lnlike={"LogLike": rastrigin}, prior=prior, restart=True)
    else:
        print("Retrieving results from previous {} scan".format(s)) 
    results[s] = get_hdf5(settings)
        
# Plot results
# Only want to do this on one process
if rank is 0:
    fig = plt.figure(figsize=(4*Nscans,8))
    for i,(s,c) in enumerate(zip(scanners,colors)):
        x,y = results[s].get_params(["x","y"])
        ax = fig.add_subplot(2,Nscans,i+1)
        ax.set_title("{0} (N={1})".format(s,len(x)))
        ax.scatter(x,y,c=c,label=s,s=0.5)
        ax = fig.add_subplot(2,Nscans,i+1+Nscans)
        results[s].plot_profile_likelihood(ax,"x","y") 
    
    ax.legend(loc=1, frameon=True, framealpha=1, prop={'size':10}) 
    plt.tight_layout()
    fig.savefig("test_all_scanners.png")



"""Python interface to ScannerBit library, from GAMBIT project"""

# Need to do a little voodoo so that dynamically loaded shared libraries
# can dynamically load other shared libraries (the scanner plugins)
import os
import sys
import ctypes
# NOTE: moved to __init__.py
# flags = sys.getdlopenflags()
# sys.setdlopenflags(flags | ctypes.RTLD_GLOBAL)

from functools import partial
import inspect
import copy
import h5py

# Just doing this will initialise MPI, and it will automatically
# call 'finalize' upon exit. So we need do nothing except import this.
from mpi4py import MPI
MPI_rank = MPI.COMM_WORLD.Get_rank()
MPI_size = MPI.COMM_WORLD.Get_size()

# DEBUG! Only used to check that YAML options looks reasonable
import yaml

# Need to tell ScannerBit where its config files are located
# We do this via a special environment variable
# Note: moved to __init__.py
# gambit_path = os.path.dirname(__file__)
# print("Setting GAMBIT_RUN_DIR to:",gambit_path)
# os.environ["GAMBIT_RUN_DIR"] = gambit_path

# Other python helper tools
from .defaults import _default_options
from .utils import _merge
from .hdf5_help import get_data, HDF5 
from .processify import processify

class SanityCheckException(Exception):
    """Exception thrown by sanity checks of user-supplied input"""
    pass

def _add_default_options(options):
   """Inspect user-supplied options, and fill in any missing
      bits with defaults"""
   _merge(options,_default_options)
   return options

# Better version of 'partial'. From 'funcy' library: https://github.com/Suor/funcy
def func_partial(func, *args, **kwargs):
    """A functools.partial alternative, which returns a real function.
       Can be used to construct methods."""
    return lambda *a, **kw: func(*(args + a), **dict(kwargs, **kw))

class SafeVec:
   """A simple wrapper class for the 'vec' argument to user prior functions,
      with nicer error messages"""
   def __init__(self,vec,size):
       self.vec = vec
       self.size = size

   def __getitem__(self, key):
       try:
           if float(key).is_integer():
             if key >= self.size:
                msg="Error accessing item '{0}' in 'vec' argument of user-supplied prior function! The size of 'vec' is {1} so entry {0} doesn't exist!".format(key,self.size) 
                raise SanityCheckException(msg)
             elif key < 0:
                msg="Error accessing item '{0}' in 'vec' argument of user-supplied prior function! The 'vec' behaves like a list and must be indexed by a non-negative integer.".format(key) 
                raise SanityCheckException(msg)
           else:
             msg="Error accessing item '{0}' in 'vec' argument of user-supplied prior function! 'vec' behaves like a list and must indexed by an integer.".format(key)
             raise SanityCheckException(msg)
       except (ValueError, AttributeError):
          msg="Error accessing item '{0}' in 'vec' argument of user-supplied prior function! '{0}' does not seem to be an integer, or even a number!".format(key)
          raise SanityCheckException(msg)
       return self.vec[key]

   def __setitem__(self, key, value):
       msg="Error setting item '{0}' to '{1}' in 'vec' argument of user-supplied prior function! 'vec' is read-only!".format(key,value) 
       raise SanityCheckException(msg)

@processify # Run this function in a separate process, so that scanner plugins can be re-loaded between scans
def _run_scan(settings, loglike_func, prior_func):
   """Perform a scan. This function is decorated in such a 
      way that it runs in a new process. This is important
      because the GAMBIT plugins can only run once per
      process, because the shared libraries need to be reloaded
      to perform a second scan.
      """
   # Import functions from the ScannerBit.so library
   # Should also trigger the loading of the scanner plugin libraries
   from . import ext_module as ext
   # check if hdf5 libraries are the same
   ext.sb.check_hdf5_version()
   # Check that ScannerBit was compiled with MPI enabled if we are using more than one process
   if MPI_size>1 and not ext.sb.WITH_MPI:
       msg = "ScannerBit has not been compiled with MPI enabled! Please try again using only one process."
       raise RuntimeError(msg)

   print("prior_func:",prior_func)
   print("prior_func:",inspect.signature(prior_func))

   # Attach the ScannerBit object to the first argument of the wrapped likelihood and prior functions
   wrapped_loglike = func_partial(loglike_func,ext.sb)
   if prior_func is not None:
       wrapped_prior = func_partial(prior_func,ext.sb)
   else:
       wrapped_prior = None

   # Create scan object
   myscan = ext.sb.scan(False) # Do not Init_MPI in ScannerBit, we already did it out here in the wrapper.

   # Double check settings fed to ScannerBit!
   print("Scan settings:")
   print(settings)
   print("==============")
   print(yaml.dump(settings, default_flow_style=False))

   # run scan
   # 'inifile' can be the name of a YAML file, or a dict.
   #myscan.run(inifile=settings, lnlike={"LogLike": wrapped_loglike}, prior=prior_func, restart=True)
   ret = myscan.run(inifile=settings, lnlike={"LogLike": wrapped_loglike}, prior=wrapped_prior, restart=True)
   
   if ret!=0:
      msg = "Fatal error encountered while running ScannerBit!"
      raise RuntimeError(msg)

   MPI.COMM_WORLD.Barrier()
   rank = MPI.COMM_WORLD.Get_rank()
   print("Rank {0} passed scan end barrier!".format(rank))

class Scan:
    """Helper object for setting up and running a scan, and
       making some basic plots.
    """
    def __init__(self, function, prior_func=None, bounds=None, prior_types=None, kwargs=None, scanner=None,
      scanner_options={}, model_name=None, output_path=None, fargs=None):
        """
        function - Python function to be scanned
        prior_func - User-define prior transformation function (optional)
        bounds - list of ranges of parameter values (function arguments) to scan,
                 or mean/std-dev in case of normal prior.
        prior_types - list of priors to use for scanning in each dimension (e.g. flat/log).
                      If None then the user is expected to do the inverse transform from
                      a unit hypercube to their parameter space themselves.
        scanner - Scanning algorithm to use
        scanner_options - Configuration options dictionary for chosen scanning algorithm
        f_args - List of names of function argments to use (if None these are
         inferred from the signature of 'function')
        """
        self.function = function
        self.prior_func = prior_func

        # Determine parameter names, either automatically or from 'fargs' argument
        if fargs is None:
            signature = inspect.signature(self.function)
            print("signature:", signature.parameters.items())
            # First argument needs to be the 'scan' object, to allow access to printers etc. Skip it in determination of parameter names.
            par_names = signature.parameters.keys()
            self._argument_names = list(par_names)[1:]
            if list(par_names)[0] is not "scan":
                msg="\n\nError in signature of user-supplied log-likelihood function! The first argument must be named 'scan' and will be passed a reference to a global ScannerBit object that can be used to e.g. print extra information to the scan output file.\n"
                raise SanityCheckException(msg)
        else:
            self._argument_names = fargs

        # Determine prior to be used
        self.bounds = bounds if bounds else [(0,1)] * len(self._argument_names)
        self.prior_types = prior_types if prior_types else ["flat"] * len(self._argument_names)
        self.scanner = scanner

        if scanner not in _default_options["Scanner"]["scanners"].keys():
            msg = "Unknown scanner '{0}' was selected! Please choose from the following available scanning algorithms:".format(scanner)
            for s in _default_options["Scanner"]["scanners"].keys():
                msg += "\n   {0}".format(s)
            raise ValueError(msg)

        if MPI_size>1 and scanner in ["random","toy_mcmc"]:
            msg = "Scanner {0} selected, however MPI_size>1, and unfortunately this algorithm is not yet parallelised. Please either choose another sampling algorithm, or run this algorithm on one process only."
            raise ValueError(msg)

        # Copy user-supplied scanner options into full scan settings dictionary
        self.settings = _add_default_options(copy.deepcopy({"Scanner": {"scanners": {scanner: scanner_options}}}))
        self.kwargs = kwargs

        print("self._argument_names:", self._argument_names)
        print("self.bounds:", self.bounds)
        assert len(self._argument_names) == len(self.bounds)
 
        if model_name is None:
           if "Parameters" in self.settings:
              #print(self.settings["Parameters"])
              self._model_name = list(self.settings["Parameters"].keys())[0]
           else:
              self._model_name = "default"
        else:
           self._model_name = model_name

        # Wrap user-supplied likelihood and prior functions with some extra sanity checking
        self._wrapped_function = self._wrap_function()
        if prior_func is not None:
           self._wrapped_prior = self._wrap_prior()
        else:
           self._wrapped_prior = None

        self._scanned = False

        # Make up a name for the run, if user didn't provide one
        if output_path is None:
            self.settings["KeyValues"]["default_output_path"] += "{0}_scan".format(self.scanner)
        else:
            self.settings["KeyValues"]["default_output_path"] = output_path

        self._process_settings()
        self.loglike_par = "LogLike"
        self.scanner = self.settings["Scanner"]["use_scanner"] # Might have been None and then filled by _process_settings with defaults
        if self.scanner == "multinest" \
        or self.scanner == "polychord":
            self.posterior_par = "Posterior"
        elif self.scanner == "twalk":
            self.posterior_par = "mult"
        else:
            self.posterior_par = None

    def _get_hdf5_group(self):
        """
        """
        assert self.settings["Printer"]["printer"] == "hdf5"
        file_name = self.settings["Printer"]["options"]["output_file"]
        group_name = self.settings["Printer"]["options"]["group"]
        DIR = self.settings["KeyValues"]["default_output_path"]
        fullpath = "{}/samples/{}".format(DIR, file_name)
        f = h5py.File(fullpath,'r')
        g = f[group_name]
        return g

    def _process_settings(self):
        """Copy 'pythonic' options into the settings dictionary
           so that they can be automaticall converted into YAML 
           by the API, which ScannerBit can then read"""
        if(self.scanner is not None):
            self.settings["Scanner"]["use_scanner"] = self.scanner
        if("Parameters" not in self.settings):
            self.settings["Parameters"] = dict()
            self.settings["Parameters"][self._model_name] = dict()
            for n in self._argument_names:
                # Just add parameter names. We will write the prior settings out fully
                # in the Priors section next
                self.settings["Parameters"][self._model_name][n] = None
        if("Priors" not in self.settings):
            self.settings["Priors"] = dict()
            if (self.prior_types is not None) \
               and (self.bounds is not None):
                for n, b, t in zip(self._argument_names, self.bounds, self.prior_types):
                    prior_setup = {'prior_type': t, 'parameters': ["{0}::{1}".format(self._model_name,n)]}
                    if t is 'gaussian' or t is 'cauchy':
                       prior_setup['mean'] = [b[0]]
                       prior_setup['cov']  = [b[1]] # Assume this is the VARIANCE, not standard deviation
                    else:
                       prior_setup['range'] = b
                    self.settings["Priors"]["{0}_prior".format(n)] = prior_setup
            else:
                raise ValueError("No prior settings found! These need to be either supplied in simplified form via the 'bounds' and 'prior_types' arguments, or else supplied in long form (following the GAMBIT YAML format) in the 'settings' dictionary under the 'Priors' key (or under the 'Parameters' key in the short-cut format)")
        #print(self.settings)
        print("Scan settings in:")
        print("==============")
        print(yaml.dump(self.settings, default_flow_style=False))

    def _wrap_function(self):
        """
        """
        def wrapped_function(scan,par_dict):
            """
            """
            arguments = [par_dict["{0}::{1}".format(self._model_name, n)]
              for n in self._argument_names]
            return self.function(scan, *arguments, **(self.kwargs or {}))

        return wrapped_function

    def _wrap_prior(self):
       """Wrap the user-supplied prior transformation function so that we can
          do some sanity checking on it"""
       def wrapped_prior(scan,vec,map):
          #Tell ScannerBit the dimension of the parameter space
          size=len(self._argument_names)
          scan.ensure_size(vec,size)
          tmp_map = {}
          self.prior_func(SafeVec(vec,size),tmp_map)
          # Check that user added all the parameters
          has_model_name=None
          for p in self._argument_names:
             if self._model_name+"::"+p not in tmp_map.keys() \
               and p in tmp_map.keys():
                if has_model_name is None: 
                   has_model_name=False
                elif has_model_name is True:
                   msg="Error in user-supplied prior transformation function! Parameter names set in 'map' are inconsistent; some appear to have the model name prefix, while others do not. Please edit your prior function to use only one format or the other."  
                   raise SanityCheckException(msg)
             elif p not in tmp_map.keys():
                   msg="Error in user-supplied prior transformation function! User must define parameters {0} or {1} in the 'map' argument, however parameter {2} was not found! Please fix your prior transformation function.".format(self._argument_names,[self._model_name+"::"+x for x in self._argument_names],p)
                   raise SanityCheckException(msg)
             else:
                if has_model_name is None: 
                   has_model_name=True
                elif has_model_name is False:
                   msg="Error in user-supplied prior transformation function! Parameter names set in 'map' are inconsistent; some appear to have the model name prefix, while others do not. Please edit your prior function to use only one format or the other."
                   raise SanityCheckException(msg)
          if not has_model_name:
             # If they added them without the 'model' part of the name then add that automatically now
             for k,v in tmp_map.items():
                 map[self._model_name+"::"+k] = v
          else:
             for k,v in tmp_map.items():
                 map[k] = v
       return wrapped_prior

    def scan(self):
       """Perform a scan. This runs a function that is decorated in such a 
       way that it runs in a new process. This is important
       because the GAMBIT plugins can only run once per
       process, because the shared libraries need to be reloaded
       to perform a second scan.

       Downside is that all arguments must be pickle-able.
       """
       try:
           _run_scan(self.settings, self._wrapped_function, self._wrapped_prior)
       except RuntimeError as err:
           # Error messages thrown in the subprocess get kind of mangled, need to
           # help it print correctly
           print("Error thrown from ScannerBit subprocess:\n")
           print(err.args[0].replace('\\n', '\n')) # Newlines appear to be weirdly escaped. Fix them. 
           exit()

       self._scanned = True

    def get_hdf5(self):
        try:
            g = HDF5(self._get_hdf5_group().id,model=self._model_name,
                loglike=self.loglike_par, posterior=self.posterior_par)
        except:
            if(self._scanned):
                raise IOError("Failed to open HDF5 output of scan!")
            else:
                raise IOError("Failed to open HDF5 output of scan, however we did not perform a scan just now. The output will only exist if you have previously run this scan. Please check that you did this!")
        return g

    def rm_samples(self):
        file_name = self.settings["Printer"]["options"]["output_file"]
        DIR = self.settings["KeyValues"]["default_output_path"]
        fullpath = "{}/samples/{}".format(DIR, file_name)
        try:
            os.remove(fullpath)
        except:
            print("No such file: " + fullpath)




import numpy as np
import h5py
import time
import csv
import os
import subprocess as sp
import array
import re
import unicodedata
import operator
import shlex

import matplotlib.pyplot as plt
import matplotlib.axes as axes
import matplotlib.cm as cm
import matplotlib.colors as colors
import matplotlib.colorbar as colorbar
import matplotlib as mpl

import scipy.stats as sps

from itertools import groupby
from operator import itemgetter

#global options that may be desired to be tweaked
font = {'family' : 'serif',
        'weight' : 'normal',  #'bold'
        'size'   : 16,
        }
mpl.rc('font', **font)
#mpl.rc('text', usetex=True)

#========Side tools===============================
#mangle unicode strings into ascii strings
def unicode2ascii(string):
    try:
        outstr = unicodedata.normalize('NFKD', string).encode('ascii','ignore')
    except TypeError:
        #If not a unicode string, do nothing.
        outstr = string       
    return outstr
 
#extract column from 2D nested list/array
def column(matrix, i):
    return [row[i] for row in matrix]
    
# uncertainty to string
from math import floor, log10
def un2str(x, xe, precision=2):
    """pretty print nominal value and uncertainty

    x  - nominal value
    xe - uncertainty
    precision - number of significant digits in uncertainty

    returns shortest string representation of `x +- xe` either as
        x.xx(ee)e+xx
    or as
        xxx.xx(ee)"""
    # base 10 exponents
    #print x, xe
    try:
        x_exp = int(floor(log10(x)))
    except ValueError:  #we get this if x is negative
        x_exp = -int(floor(log10(-x)))  #MIGHT NEED TO CHANGE THIS TO CEIL? HAVE NOT CHECKED THOROUGHLY
    try:
        xe_exp = int(floor(log10(xe)))
    except ValueError:
        xe_exp = -int(floor(log10(-xe)))

    # uncertainty
    un_exp = xe_exp-precision+1
    un_int = round(xe*10**(-un_exp))

    # nominal value
    no_exp = un_exp
    no_int = round(x*10**(-no_exp))

    # format - nom(unc)exp
    fieldw = - x_exp - no_exp
    fmt = '%%.%df' % fieldw
    result1 = (fmt + '(%.0f)e%d') % (no_int*10**(-fieldw), un_int, x_exp)
    # format - nom(unc)
    fieldw = max(0, -no_exp)
    fmt = '%%.%df' % fieldw
    result2 = (fmt + '(%.0f)') % (no_int*10**no_exp, un_int*10**max(0, un_exp))

    # return shortest representation
    if len(result2) <= len(result1):
        return result2
    else:
        return result1
        
        
       
#==============ANALYSIS FUNCTIONS=======================================

#custom colormap for Delta chi^2 plots

#Define contours of likelihood function to plot (in sigma units)
#contsigmas = [0.5,1,2,3,4,5,6,7]
#contsigmas = [1,2]    #reduced set to remove clutter
#rellevels = np.array(contsigmas)**2 #square every element to get them into chi^2 units

#compute delta chi^2 for 68 and 95% 2D confidence regions 
CLs = [0.68,0.95]
df = 2  #two degrees of freedom in 2D profile likelihood plot
rellevels = [sps.chi2.isf(1-CL,df) for CL in CLs]   #isf = inverse survival function (inverse of 1-cdf). returns chi2 value at boundary of specified confidence region

#Define colormap for chi^2 plots
#scaled colors go from 0 to 1, so we need to set a max value (chi^2=25 sounds good)
#actually plot "sigmas", i.e. sqrt(chi^2), so max is then 5.
mn=0
mx=5
s0=0./(mx-mn)
s1=1./(mx-mn)
s2=2./(mx-mn)
s3=3./(mx-mn)
s4=4./(mx-mn)
s5=5./(mx-mn)

#colors are in order ('green':(0., 1., 0.), 'yellow': (1., 1., 0.), 'orange' : (0., 0.5, 1.)
# 'red' : (0., 0., 1.), 'darkred' : (0., 0., 0.5), 'darkerred' : (0., 0., 0.2) (names just made up)
#note, the rgb values are a bit screwed up in the previous list, but they are correct in the cdict
cdict = {
'red'  :  ((s0, 0., 0.), (s1, 1., 1.), (s2, 1., 1.), (s3, 1., 1.), (s4, .5, .5), (s5, .2, .2)),
'green':  ((s0, 1., 1.), (s1, 1., 1.), (s2, .5, .5), (s3, 0., 0.), (s4, 0., 0.), (s5, 0., 0.)),
'blue' :  ((s0, 0., 0.), (s1, 0., 0.), (s2, 0., 0.), (s3, 0., 0.), (s4, 0., 0.), (s5, 0., 0.))
}

#generate the colormap with 1024 interpolated values
chi2cmap = colors.LinearSegmentedColormap('chi2_colormap', cdict, 1024)
chi2cmap.set_bad('w',1.)    #set color (and alpha) for bad ('masked') values

#cutsom colormap for marginalised posterior plots

cdict2 = {
'red'  :  ((0., 1., 1.), (1., 0., 0.)),
'green':  ((0., 1., 1.), (1., 0., 0.)),
'blue' :  ((0., 1., 1.), (1., 0., 0.))
}

cdict3 = {
'red'  :  ((0., 0., 0.), (1., 1., 1.)),
'green':  ((0., 0., 0.), (1., 1., 1.)),
'blue' :  ((0., 0., 0.), (1., 1., 1.))
}

cdict4 = {
'red'  :  ((0., 1., 1.), (.01, 0., 0.), (.05, .1, .1), (1., .9, .9)),
'green':  ((0., 1., 1.), (.01, .3, .3), (.05, .1, .1), (1., .9, .9)),
'blue' :  ((0., 1., 1.), (.01, .3, .3), (.05, .1, .1), (1., 1., 1.))
}

#generate the colormap with 1024 interpolated values
margcmap = colors.LinearSegmentedColormap('marg_colormap', cdict4, 1024)
margconts = [0.68,0.95]

def getcols(structarr,colnames):
    """"helper function to retrieve a normal, unstructured numpy array from the
    structured array that the dataset returns
    Args
    structarr - result of dset[:], or dset[0:1000] if a subset of rows is desired
    colnames  - list of field names to retrieve from array
    """
    """ This way may get the columns switched around!
    #As a bonus, remove NaN's from data
    
    data = structarr[:][colnames].view(dtype='<f4').reshape(-1,len(colnames))
    dataout = data[~np.isnan(data).any(1)]  #deletes rows containing nans
    print 'NaN count: ',len(data) - len(dataout)
    """
    #Pull columns out one at a time to ensure we get the correct order
    data = np.array([structarr[col] for col in colnames]).transpose()
    print(data.shape, type(data))
    try:
        dataout = data[~np.isnan(data).any(1)]  #deletes rows containing nans
    except TypeError as err:
        print('TypeError encountered, dumping extra data...')
        for i,row in enumerate(data):
            try:
                ~np.isnan(row).any(1)
            except (TypeError,NotImplementedError) as err2:
                print('TypeError or NotImplementedError occurred running isnan \
function on row {0}, dumping row...'.format(i))
                print(row, type(row))
                raise err2
            
    print(dataout.shape)
    print('NaN count: ',len(data) - len(dataout))
    return dataout

def bin2d(data,nxbins,nybins,binop='min',doconts=False):
    """2d binning algorithm
    Args
    data - n rows * 3 column dataset, columns in order x,y,z. Binning done
    by x,y
    binop - Operation to perform on bins. Available options:
        'min' - Returns the minimum z value for each bin
        'max' - Returns the maximum z value for each bin
    nxbins - Number of bins to use in x direction
    nybins -    "      "              y   "
    """
    x = data[:,0]
    y = data[:,1]
    #xbins = np.arange(min(x),
    zvals = data[:,2]
    eps = 1e-10     #shift for bin edge computation

    xindexes = np.floor((1-eps)*(nxbins)*(x-min(x))/(max(x)-min(x)))
    yindexes = np.floor((1-eps)*(nybins)*(y-min(y))/(max(y)-min(y)))
    #print xindexes
    #print yindexes
    
    outarray = np.zeros((nxbins,nybins))
    #print outarray
    
    #grouped=[list(value) for key, value in groupby(sorted(zip(xindexes,yindexes,zvals)), key=itemgetter(0,1))]
    grouped=[list(value) for key, value in groupby(sorted(zip(xindexes,yindexes,zvals)), key=itemgetter(0,1))]
    bininds=[nybins*bin[0][0] + bin[0][1] for bin in grouped]  #these are the FLAT indices in the target array
    #print("bininds:",bininds)
    if binop=='min':
        binvals=[min([el[2] for el in bin]) for bin in grouped]
        #outarray[:,:] = 1e300
        outarray[:,:] = np.nan
    elif binop=='max':
        binvals=[max([el[2] for el in bin]) for bin in grouped]
        outarray[:,:] = 0
    elif binop=='sum':
        binvals=[sum([el[2] for el in bin]) for bin in grouped]
        outarray[:,:] = 0
    else:
        raise ValueError('Invalid binop value ({0}) supplied to bin2d'.format(binop))

    #print bininds
    try:
        outarray.flat[bininds] = binvals
    except ValueError:
        print("WARNING: nan's present in bin index lists, input data may\
be degenerate in one or more dimensions.")
        raise  
    #Recompute the bins so that each one contains the total probability enclosed by the
    #iso-probability-density contour on which it sits (use this to compute smallest 68%, 95% 
    #Bayesian credible regions) (DESIGNED FOR USE WHEN BIN VALUE IS A PROBABILITY MASS)
    if binop=='sum' and doconts==True:
        outarray2 = np.zeros((nxbins,nybins))
        outarray2[:,:] = 1.
        sb = np.array(sorted(zip(bininds,binvals), key=itemgetter(1), reverse=True)) #sort bins by probability
        #print zip(sb[:,1],np.cumsum(sb[:,1]))[0:20]
        #print zip(sb[:,1],np.cumsum(sb[:,1]))[-20:-1]
        outarray2.flat[list(sb[:,0])] = list(np.cumsum(sb[:,1]))         #cumulative sum of probabilities
        return outarray.transpose(), outarray2.transpose()
        
    return outarray.transpose()
    
def bin1d(data,nbins,binop='min'):
    """1d version of bin2d (in fact uses bin2d directly)
    Args:
    data - n rows * 2 column dataset, columns in order x,z. Binning done
    in x, z being the "height" or "density" to be binned.
    binop - Operation to perform on bins. Available options:
        'min' - Returns the minimum z value for each bin
        'max' - Returns the maximum z value for each bin
    nbins - Number of bins to use 
    """
    x = data[:,0]
    y = np.zeros(x.shape[0]) #"fake" y values to emulate the 2D binning
    y[:x.shape[0]/2] = 1  #need at least two different y values so ymax - ymin != 0.
    z = data[:,1]
    return bin2d(np.array(zip(x,y,z)),nbins,1,binop,doconts=False)[0]      #return 1d array of x vs binned values

def forceAspect(ax,aspect=1):
    extent = ax.get_window_extent().get_points()
    print(extent)
    ax.set_aspect(abs((extent[0,1]-extent[0,0])/(extent[1,1]-extent[1,0]))/aspect)
    
    """
    print ax
    print ax.get_images()
    if isinstance(ax, axes.Axes):
        extent =  ax.get_extent()
    elif isinstance(ax.get_images()[0], plt.axes.Axes):
        im = ax.get_images()
        extent =  im[0].get_extent()
    else:
        raise TypeError("Supplied 'ax' was not an Axes object or list of Axes objects!")
    ax.set_aspect(abs((extent[1]-extent[0])/(extent[3]-extent[2]))/aspect)
    """
    
#--------------main plotting routines-----------------------------------

def chi2scatplot(ax,data,title=None,labels=None,s=1):
    """Creates a scatter plot of the data, colored by Delta chi^2 value
    Args:
    ax - Axis object on which to create plot
    data - 3 column numpy array with
        data[:,0] - x data
        data[:,1] - y data
        data[:,2] - chi^2 data
    """
    data = data[data[:,2].argsort()[::-1]] #sort points by chi2 (want to plot lowest chi2 points last, achieved by reversing sorted indices via '::-1')
    plot = ax.scatter(data[:,0],data[:,1],c=np.sqrt(data[:,2]-min(data[:,2])),s=s,lw=0,cmap=chi2cmap, norm=colors.Normalize(vmin=mn,vmax=mx,clip=True))
    if title: ax.set_title(title)
    if labels: ax.set_xlabel(labels[0])
    if labels: ax.set_ylabel(labels[1])
    ax.set_xlim(min(data[:,0]),max(data[:,0]))
    ax.set_ylim(min(data[:,1]),max(data[:,1]))
    ax.grid(True)
    return plot

def chi2logscatplot(ax,data,title=None,labels=None):
    """Creates a scatter plot of the data, colored by Delta chi^2 value
    """
    data = data[data[:,2].argsort()[::-1]] #sort points by chi2 (want to plot lowest chi2 points last, achieved by reversing sorted indices via '::-1')
    plot = ax.scatter(data[:,0],data[:,1],c=np.sqrt(data[:,2]-min(data[:,2])),s=1,lw=0,cmap=chi2cmap, norm=colors.Normalize(vmin=mn,vmax=mx,clip=True))
    ax.set_yscale('log')
    if title: ax.set_title(title)
    if labels: ax.set_xlabel(labels[0])
    if labels: ax.set_ylabel(labels[1])
    ax.set_xlim(min(data[:,0]),max(data[:,0]))
    ax.set_ylim(min(data[:,1]),max(data[:,1]))
    ax.grid(True)
    return plot
    
def profplot(ax,data,title=None,labels=None,nybins=100,nxbins=None):
    """Creates a binned, profiled plot of the data, colored by Delta chi^2 value,
    i.e. profile likelihood.
    """
    if nxbins is None:
        nxbins=np.floor(1.618*nybins)
    x = data[:,0]
    y = data[:,1]
    wx= (max(x)-min(x))/nxbins
    wy= (max(y)-min(y))/nybins
    #print(min(x),max(x),wx)
    #print(min(y),max(y),wy)
    xlist = np.arange(min(x),max(x)-wx*10e-3,wx)   #tiny shift to prevent nbins+1 bins in X and Y arrays
    ylist = np.arange(min(y),max(y)-wy*10e-3,wy)
    #print(len(xlist), len(ylist))
    
    outarray = bin2d(data,nxbins,nybins,binop='min')
    #print outarray[0:20,0:20]
    X, Y = np.meshgrid(xlist,ylist)
    minchi2 = np.nanmin(outarray)  #get min ignoring nans
    #print X.shape, Y.shape, outarray.shape, minchi2
    
    #Set special color for NaN values (see definition of colormap also)
    Dchi2 = np.sqrt(outarray - minchi2)
    #print Dchi2[0:20,0:20]

    masked_array = np.ma.array(Dchi2, mask=np.isnan(Dchi2)) #mask out the nan's; the masked values will be given a special color in the output plot
    
    #print masked_array[0:20,0:20]
    
    im = ax.imshow(masked_array, origin='lower', interpolation='nearest',
                    extent=(min(xlist),max(xlist)+wx,min(ylist),max(ylist)+wy),
                    cmap=chi2cmap, norm=colors.Normalize(vmin=mn,vmax=mx,clip=True), aspect='auto')
    
    CS = ax.contour(X+wx/2, Y+wy/2, outarray - minchi2, levels=rellevels, lw=3) #colors=['g','y','r']
    bfidx=np.where(outarray.flat==minchi2) #get flat index of best fit point
    ax.plot([X.flat[bfidx]], [Y.flat[bfidx]], "ko")
    #ax.clabel(CS, inline=1, fontsize=10)
    
    if labels: ax.set_xlabel(labels[0])
    if labels: ax.set_ylabel(labels[1])
    ax.set_xlim(min(data[:,0]),max(data[:,0]))
    ax.set_ylim(min(data[:,1]),max(data[:,1]))
    ax.grid(True)
    
    return im
    
def margplot(ax,data,nxbins=100,nybins=100,title=None,labels=None):
    """Creates a binned marginalised plot of the data, colored by marginalised posterior
    density.
    """
    x = data[:,0]
    y = data[:,1]
    wx= (max(x)-min(x))/nxbins
    wy= (max(y)-min(y))/nybins
    xlist = np.arange(min(x),max(x)-wx*10e-3,wx)    #tiny shift to prevent nbins+1 bins in X and Y arrays
    ylist = np.arange(min(y),max(y)-wy*10e-3,wy)
    
    outarraydens, outarrayconts = bin2d(data,nxbins,nybins,binop='sum',doconts=True)
    X, Y = np.meshgrid(xlist,ylist)
    maxpoint = max(outarraydens.flat)

    #print(X.shape, Y.shape, outarraydens.shape, outarrayconts.shape, max(outarrayconts.flat))
    
    im = ax.imshow(outarraydens, origin='lower', interpolation='nearest',
                    extent=(min(xlist),max(xlist)+wx,min(ylist),max(ylist)+wy),
                    cmap=margcmap, aspect='auto')
    
    CS = ax.contour(X+wx/2, Y+wy/2, outarrayconts, levels=margconts, lw=3, colors=[(0,1,0),(1,1,0)])    #(r,g,b)
    bfidx=np.where(outarraydens.flatten()==maxpoint) #get flat index of best fit point
    ax.plot([X.flat[bfidx]], [Y.flat[bfidx]], "ko")
    #ax.clabel(CS, inline=1, fontsize=10)
    
    if labels: ax.set_xlabel(labels[0])
    if labels: ax.set_ylabel(labels[1])
    ax.set_xlim(min(data[:,0]),max(data[:,0]))
    ax.set_ylim(min(data[:,1]),max(data[:,1]))
    ax.grid(True)
    
    return im

def margplot1D(ax,data,title=None,labels=None,trim=True):
    """Creates a 1D binned marginalised plot of the data
    Args:
    data[:,0] - x data
    data[:,1] - corresponding posterior masses 
    trim - Cut off outer bins containing less than 0.01% probability mass
    """
    nxbins=200

    x = data[:,0]
    wx= (max(x)-min(x))/nxbins
    xlist = np.arange(min(x),max(x)-wx*10e-3,wx)    #tiny shift to prevent nbins+1 bins in X and Y arrays
    
    outarraydens = bin1d(data,nxbins,binop='sum')
    
    # Perform trimming:
    # Do binning with all the data, cut off bins that contain almost no 
    # probability, then repeat binning over reduced range (iterate)
    if trim:
        while True:
            # from low
            posl = 0
            pmass = outarraydens[posl]
            while pmass<0.001:
                posl += 1
                pmass = outarraydens[posl]
            # from high
            posh = len(outarraydens)-1
            pmass = outarraydens[posh]
            while pmass<0.001:
                posh -= 1
                pmass = outarraydens[posh]
            # compute new range and repeat binning
            data = data[ np.logical_and(data[:,0]>xlist[posl], data[:,0]<xlist[posh+1]) ] 
            x = data[:,0]
            wx= (max(x)-min(x))/nxbins
            xlist = np.arange(min(x),max(x)-wx*10e-3,wx)    # tiny shift to prevent nbins+1 bins in X and Y arrays
            if posh-posl > 80:  # if we aren't cutting off many bins don't bother to optimise further
                break
            # repeat binning
            outarraydens = bin1d(data,nxbins,binop='sum')

    #im = ax.plot(xlist,outarraydens)   
    im = ax.bar(xlist,outarraydens,wx,linewidth=0,color='b',alpha=0.4)
      
    if labels: ax.set_xlabel(labels)
    ax.set_ylabel("Binned posterior mass")
    ax.set_xlim(min(data[:,0]),max(data[:,0]))
    
    return im




"""Demo script which runs all the (serious) scanners to which pyScannerBit has access"""

import numpy as np
import math
import pyscannerbit.scan as sb
import matplotlib.pyplot as plt

from mpi4py import MPI
rank = MPI.COMM_WORLD.Get_rank()
size = MPI.COMM_WORLD.Get_size()

# Regenerate scan data?
new_scans = True

# Test function
def rastrigin(scan,x,y,z):
    X = [x,y,z]
    A = 10
    return - (A + sum([(x**2 - A * np.cos(2 * math.pi * x)) for x in X]))

# Prior transformation from unit hypercube
def prior(vec, map):
    map["x"] = -4 + 8*vec[0] # flat prior over [-4,4]
    map["y"] = -4 + 8*vec[1]
    map["z"] = -4 + 8*vec[2]

# Settings for quick and dirty scans. Won't do very well, because the test function is
# actually rather tough!
# Don't have to specify all scanner options; anything missing will revert to defaults (see defaults.py)
scanner_options = {}
scanner_options["multinest"] = {"tol": 0.5, "nlive": 100} 
scanner_options["polychord"] = {"tol": 1.0, "nlive": 20} 
scanner_options["diver"]     = {"convthresh": 1e-2, "NP": 300} 
scanner_options["twalk"]     = {"sqrtR": 1.05}
scanner_options["random"]    = {"point_number": 10000}
scanner_options["toy_mcmc"]  = {"point_number": 10} # Acceptance ratio is really bad with this scanner, so don't ask for much
scanner_options["badass"  ]  = {"points": 1000, "jumps": 10}
scanner_options["pso"]       = {"NP": 400}

scanners = ["multinest","polychord","diver","twalk"]
#colors = ["r","m","b","g"]
#scanners = ["twalk","badass","pso"]
colors = ["r","b","g"]
if size is 1:
    scanners += ["random","toy_mcmc"] # "random" and "toy_mcmc" seem to not be MPI compatible. Should make GAMBIT throw an error about this, or fix them.
    colors += ["c","y"]

# Test just one scanner
#scanners = ["pso"]
#colors = ["r"]

Nscans = len(scanners)
results = {}

# Do all scans
for s in scanners:
    # Create scan manager object
    # (prior_types argument currently does nothing)
    myscan = sb.Scan(rastrigin, prior_func=prior, scanner=s, scanner_options=scanner_options[s])
    myscan.rm_samples()
    if new_scans:
        print("Running scan with {}".format(s))
        myscan.scan()
    else:
        print("Retrieving results from previous {} scan".format(s)) 
    results[s] = myscan.get_hdf5()
        
# Plot results
# Only want to do this on one process
if rank is 0:
    fig = plt.figure(figsize=(4*Nscans,8))
    for i,(s,c) in enumerate(zip(scanners,colors)):
        x,y = results[s].get_params(["x","y"])
        ax = fig.add_subplot(2,Nscans,i+1)
        ax.set_title("{0} (N={1})".format(s,len(x)))
        ax.scatter(x,y,c=c,label=s,s=0.5)
        ax = fig.add_subplot(2,Nscans,i+1+Nscans)
        results[s].plot_profile_likelihood(ax,"x","y") 
    
    ax.legend(loc=1, frameon=True, framealpha=1, prop={'size':10}) 
    plt.tight_layout()
    fig.savefig("test_all_scanners.png")



import os
import re
import sys
import sysconfig
import site
import platform
import subprocess
try:
    import pathlib
except ImportError:
    raise ImportError("pyScannerBit requires the 'pathlib' package for installation. Please install it and try again.")

import multiprocessing

from distutils.version import LooseVersion
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext as build_ext_orig

class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=''):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)

class CMakeBuild(build_ext_orig):
    def run(self):
        try:
            out = subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError("CMake must be installed to build the following extensions: " +
                               ", ".join(e.name for e in self.extensions))

        if platform.system() == "Windows":
            raise RuntimeError("Sorry, pyScannerBit doesn't work on Windows platforms. Please use Linux or OSX.")

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        cwd = pathlib.Path().absolute()

        # these dirs will be created in build_py, so if you don't have
        # any python sources to bundle, the dirs will be missing
        build_temp = pathlib.Path(self.build_temp)
        try:
           build_temp.mkdir(parents=True)
        except FileExistsError:
           pass # This is not a problem, but exist_ok argument doesn't exist in Python <3.5
        extdir = pathlib.Path(self.get_ext_fullpath(ext.name))
        try:
           extdir.mkdir(parents=True)
        except FileExistsError:
           pass

        # Temporary directory where libraries and other built files should go
        # These will get automatically copied to the final install location
        print('extdir:', extdir)
        #quit()
        libout = str(extdir.parent.absolute()) + '/pyscannerbit'

        # Set "ORIGIN" variable depending on OS
        if    platform.system() == "Linux":
            origin = r"$ORIGIN"
        elif  platform.system() == "Darwin": 
            # This is any OSX-like system
            origin = r"@loader_path"
        else:
            raise RuntimeError("Unrecognised operating system! pyScannerBit is only compatible with Linux and OSX-like operating systems! Aborting install...")

        cmake_args = ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + libout,
                      '-DCMAKE_VERBOSE_MAKEFILE:BOOL=OFF',
                      '-Wno-dev',
                      '-DCMAKE_RUNTIME_OUTPUT_DIRECTORY=' + libout,
                      '-DSCANNERBIT_STANDALONE=True',
                      '-DCMAKE_INSTALL_RPATH={0}'.format(origin),
                      '-DCMAKE_BUILD_WITH_INSTALL_RPATH:BOOL=ON',
                      '-DCMAKE_INSTALL_RPATH_USE_LINK_PATH:BOOL=ON',
                      '-DCMAKE_INSTALL_PREFIX:PATH=' + libout,
                      '-DWITH_MPI=ON',
                      '-DNO_PYTHON_LIBS=TRUE',
                      '-DVERBOSE_SCAN_PYTHON=TRUE',
                      '-DCMAKE_DISABLE_FIND_PACKAGE_pybind11=1',
                      '-DWITH_PYTHON_SCANNERBIT=ON',
                      #'-DCMAKE_VERBOSE_MAKEFILE=ON',
                      #'--trace-expand',
                      #'-DCMAKE_FIND_DEBUG_MODE=ON',           
                      #'-DPYBIND11_PYTHON_VERSION=3.6',
                    ]

        if sys.version_info[0] < 3:
            cmake_args += ['-DFORCE_PYTHON2=True']

        self.debug = True
        cfg = 'Debug' if self.debug else 'Release'
        build_args = ['--config', cfg]

        ncpus = multiprocessing.cpu_count()
        if ncpus>1:
           ncpus -= 1 # Use 1 fewer cpus than available, so the OS can still do other things
       
        cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]
        build_args += ['--', '-j{0}'.format(ncpus)]

        env = os.environ.copy()
        env['CXXFLAGS'] = '{} -DVERSION_INFO=\\"{}\\"'.format(env.get('CXXFLAGS', ''),
                                                              self.distribution.get_version())
       
        # # Debugging paths
        # print("Debugging paths:")
        print("extdir:",extdir)
        print("libout:",libout)
        subprocess.check_call(['pwd'], cwd=self.build_temp, env=env)
        subprocess.check_call(['echo',self.build_temp], cwd=self.build_temp, env=env)
        # # cwd = pathlib.Path().absolute()
        # # print("ext.sourcedir:")
        # # subprocess.check_call(['echo', ext.sourcedir], cwd=self.build_temp, env=env)
        # # subprocess.check_call(['echo',cwd], cwd=self.build_temp, env=env) 
        subprocess.check_call(['ls', ext.sourcedir], cwd=self.build_temp, env=env)
        subprocess.check_call(['ls', ext.sourcedir+'/pyscannerbit/scannerbit'], cwd=self.build_temp, env=env)
   
        # untar ScannerBit tarball
        subprocess.check_call(['tar','-C','pyscannerbit/scannerbit/untar/ScannerBit','-xf','pyscannerbit/scannerbit/ScannerBit_stripped.tar','--strip-components=1'], cwd=ext.sourcedir, env=env)
      
        # First cmake
        subprocess.check_call(['cmake', '-Ditch=Mathematica', ext.sourcedir] + cmake_args, cwd=str(build_temp), env=env)
        # Build all the scanners
        subprocess.check_call(['cmake', '--build', '.', '--target', 'scanners'] + build_args, cwd=str(build_temp))
        # Re-run cmake to detect built scanner plugins
        subprocess.check_call(['cmake', ext.sourcedir], cwd=str(build_temp))
        # Main build
        subprocess.check_call(['cmake', '--build', '.', '--target', 'post_build'] + build_args, cwd=str(build_temp))
        # Install
        #subprocess.check_call(['cmake', '--build', '.', '--target', 'install'], cwd=str(build_temp))

        # We need to add some __init__.py files to the package
        f=open("{0}/ScannerBit/__init__.py".format(libout),"w+")
        f.close()
        f=open("{0}/ScannerBit/python/__init__.py".format(libout),"w+")
        f.close()

        print("Checking contents of temporary directory {0}".format(libout))
        print(subprocess.check_call(['ls', libout], cwd=self.build_temp, env=env))
 
setup(
    name='pyscannerbit',
    version='0.1.1',
    author='Ben Farmer',
    # Add yourself if you contribute to this package
    author_email='ben.farmer@gmail.com',
    description='A python interface to the GAMBIT scanning module, ScannerBit',
    long_description='',
    ext_modules=[CMakeExtension('_gambit')],
    cmdclass=dict(build_ext=CMakeBuild),
    zip_safe=False,
    packages=['pyscannerbit'],
    install_requires=[
        'six',
        'pyyaml',
        'h5py',
        'mpi4py',
        'numpy',
        'scipy',
        'matplotlib',
    ],
    extras_require = {
        'sqlite3printer':  ["sqlite3"]
    }
)

# No 'package_data' stuff. We just need to get CMake to put everything into 'tmpdir' and it should get copied.



# NOTE: If you want to ScannerBit from a different directory
#       put this directory (ScannerBit/python) in the
#       python environment variable "PYTHONPATH".

import yaml
from pyscannerbit import ext_module as ext

print("WITH_MPI=",ext.sb.WITH_MPI) # Test if ScannerBit was compiled with MPI enabled

# define likelihood, technically optional
def like(m):
    a = m["model1::x"]
    ext.sb.print("my_param", 0.5) # can print custom parameters 
    
    return -a*a/2.0

# define prior, optional
def prior(vec, map):
    # tell ScannerBit that the hypergrid dimension is 1
    ext.sb.ensure_size(vec, 1) # this needs to be the first line!
    
    map["model1::x"] = 5.0 - 10.0*vec[0]

# declare scan object
myscan = ext.sb.scan(True)

# run scan from inifile
# myscan.run(inifile="ScannerBit.yaml", lnlike={"LogLike": like}, prior=prior, restart=True)

# Run scan from dict instead of YAML file
settings = {
"Parameters": {
  "model1": {
    "x": None,
    }
  },
"Priors": {
  "x_prior": {
    "prior_type": 'flat',
    "parameters": ['model1::x'],
    "range": [1.0, 40.0],
    }
  },
"Printer": {
  "printer": "hdf5",
  "options": {
    "output_file": "results.hdf5",
    "group": "/",
    "delete_file_on_restart": "true",
    }
  },
"Scanner": {
  "scanners": {
    "twalk": {
      "plugin": "twalk",
      "like": "LogLike",
      "tolerance": 1.003,
      "kwalk_ratio": 0.9,
      "projection_dimension": 4
      },
    "multinest": {
      "plugin": "multinest",
      "like": "LogLike",
      "nlive": 2000,
      "tol": 0.1
      }
    },    
  "use_scanner": "multinest",
  "objectives": {
    "gaussian": {
      "plugin": "gaussian",
      "purpose": "LogLike",
      "parameters": {
        "param...20": None,
        "range": [-5, 5]
        }
      }
    }
  },
"KeyValues": {
  "default_output_path": "pyscannerbit_run_data/",
  "likelihood": {
    "model_invalid_for_lnlike_below": -1e6
    }
  }
}

myscan.run(inifile=settings, lnlike={"LogLike": like}, prior=prior, restart=True)

###
### can run diagnostics
### NOTE: can specify any number of object to run diagnostics on
###       e.g. myscan.diagnostics("twalk", "multinest", "ais", "priors", "scanners", ...)
###
# myscan.diagnostics("twalk")

###
### can use with pyyaml!
### NOTE: yaml vector keys, [...], must be preceded by !!/python/tuple 
###
# with open("ScannerBit.yaml", 'r') as f:
#     try:
#         inifile = yaml.load(f)
#     except yaml.YAMLError as exc:
#         print(exc)
# 
# myscan.run(inifile=inifile, lnlike={"LogLike": like}, prior=prior, restart=True)

###
### or can also run multiple proposes
###
# lnlike = {
#     "LogLike1": like1
#     "LogLike2": like2
# }
# myscan.run(inifile="ScannerBit.yaml", lnlike=lnlike, prior=prior, restart=True)
